import torch
import torch.nn as nn


class Diacritics(nn.Module):
    """
    The pytorch model :class:Diacritics for diacritics restoration
        :param nr_classes: for Romanian language we defined 3 classes
                            0 - No diacritics,
                            1 - One of the following transformations
                                    "a" -> "ă"
                                    "i" -> "î"
                                    "s" -> "ș"
                                    "t" -> "ț"
                            2 - The alternative transformation for "a" -> "â"

        :param word_embedding_size: Word Embedding vector size
        :param character_embedding_size: Character Embedding vector size
        :param char_vocabulary_size: How many tokens in the word embedding vocabulary
        :param char_padding_index: How many characters in the character embedding vocabulary (can be more than 255 - UTF-8)
        :param character_window: The analysed character context window size in the text (during training)
        :param sentence_window: The analysed word context window size in the text (during training)
        :param characters_lstm_size: Hidden LSTM size for the character embedding pathway
        :param sentence_lstm_size: Hidden LSTM size for the sentence embedding pathway
        :param dropout: Dropout probability for the classification layer (relevant for train only)

    """

    def __init__(
        self,
        nr_classes=3,
        word_embedding_size=300,
        character_embedding_size=20,
        char_vocabulary_size=771,
        char_padding_index=0,
        character_window=13,
        sentence_window=31,
        characters_lstm_size=64,
        sentence_lstm_size=256,
        dropout=0.2,
    ):
        super(Diacritics, self).__init__()
        self.nr_classes = nr_classes
        self.character_embedding_size = character_embedding_size
        self.char_vocabulary_size = char_vocabulary_size
        self.word_embedding_size = word_embedding_size
        self.char_padding_index = char_padding_index
        self.characters_lstm_size = characters_lstm_size
        self.sentence_lstm_size = sentence_lstm_size
        self.character_window = character_window
        self.sentence_window = sentence_window
        self.dropout = dropout

        self.embedding = nn.Embedding(
            self.char_vocabulary_size,
            self.character_embedding_size,
            self.char_padding_index,
        )
        self.character_lstm_layer = nn.LSTM(
            input_size=self.character_embedding_size,
            hidden_size=self.characters_lstm_size,
            num_layers=1,
            batch_first=True,
            bidirectional=True,
        )

        self.sentence_bi_lstm_layer = nn.LSTM(
            input_size=self.word_embedding_size,
            hidden_size=self.sentence_lstm_size,
            num_layers=1,
            batch_first=True,
            bidirectional=True,
        )

        concat_size = (
            2 * self.characters_lstm_size
            + 2 * self.sentence_lstm_size
            + self.word_embedding_size
        )

        self.drop = nn.Dropout(p=self.dropout)

        self.dense = nn.Linear(concat_size, self.nr_classes)

    def forward(self, char_input, word_embedding, sentence_embedding):
        """
        :param char_input: shape=(self.window_character * 2 + 1,)
        :param word_embedding: shape=(word_embedding_size,)
        :param sentence_embedding: shape=(self.window_sentence * 2 + 1, self.word_embedding_size,)
        :return:
        """
        char_emb = self.embedding(char_input)
        device = next(self.parameters()).device

        h = torch.zeros(
            (
                2 * self.character_lstm_layer.num_layers,
                char_emb.size(0),
                self.character_lstm_layer.hidden_size,
            )
        )
        c = torch.zeros(
            (
                2 * self.character_lstm_layer.num_layers,
                char_emb.size(0),
                self.character_lstm_layer.hidden_size,
            )
        )
        torch.nn.init.xavier_normal_(h)
        torch.nn.init.xavier_normal_(c)

        char_hidden, _ = self.character_lstm_layer(
            char_emb, (h.to(device), c.to(device))
        )

        h = torch.zeros(
            (
                2 * self.sentence_bi_lstm_layer.num_layers,
                sentence_embedding.size(0),
                self.sentence_bi_lstm_layer.hidden_size,
            )
        )
        c = torch.zeros(
            (
                2 * self.sentence_bi_lstm_layer.num_layers,
                sentence_embedding.size(0),
                self.sentence_bi_lstm_layer.hidden_size,
            )
        )
        torch.nn.init.xavier_normal_(h)
        torch.nn.init.xavier_normal_(c)

        sentence_hidden, _ = self.sentence_bi_lstm_layer(
            sentence_embedding, (h.to(device), c.to(device))
        )

        concatenated = torch.cat(
            (char_hidden[:, -1, :], word_embedding, sentence_hidden[:, -1, :]), dim=-1
        )

        concatenated = self.drop(concatenated)
        out = self.dense(concatenated)

        # out = F.softmax(out, dim=-1)

        return out

    def save(self, filename, vocabulary, epoch=None, valid_acc=0.0, valid_f1=0.0):
        to_save = {
            "epoch": epoch,
            "model_state": self.state_dict(),
            "vocabulary": vocabulary,
            "hyperparams": {
                "nr_classes": self.nr_classes,
                "character_embedding_size": self.character_embedding_size,
                "char_vocabulary_size": self.char_vocabulary_size,
                "word_embedding_size": self.word_embedding_size,
                "char_padding_index": self.char_padding_index,
                "characters_lstm_size": self.characters_lstm_size,
                "sentence_lstm_size": self.sentence_lstm_size,
                "character_window": self.character_window,
                "sentence_window": self.sentence_window,
            },
            "valid_acc": valid_acc,
            "valid_f1": valid_f1,
        }
        torch.save(to_save, filename)



from pathlib import Path

import numpy as np
import torch
import torch.nn.functional as F
import torch.optim as optim
from sklearn.metrics import f1_score, accuracy_score, classification_report
from torch.utils.data import DataLoader
from tqdm import tqdm


def train(
    model,
    loss_func,
    train_dataloader: DataLoader,
    valid_dataloader: DataLoader,
    epochs=10,
    checkpoint_file=None,
):

    optimizer = optim.Adam(model.parameters(), lr=0.001)
    device = next(model.parameters()).device
    print(f"{device} device for training")
    # initialize running values
    history = {
        "train_loss": [],
        "valid_loss": [],
        "acc": [],
        "f1": [],
        "epoch": [],
        "step": [],
    }
    best_acc = 0
    best_acc_epoch = 0
    nr_non_improving = 0
    patience = 3

    def evaluate_step(step, epoch, running_loss, max_eval_steps=None):
        nonlocal best_acc, best_acc_epoch, nr_non_improving
        if valid_dataloader is None:
            return

        print(f"Evaluating (Epoch/Step) {epoch +1} / {step}: ")
        average_train_loss = running_loss / step
        valid_loss, valid_acc, valid_f1 = evaluate(
            model, valid_dataloader, loss_func, epoch, max_eval_steps
        )
        history["epoch"].append(epoch)
        history["step"].append(step)
        history["train_loss"].append(average_train_loss)
        history["valid_loss"].append(valid_loss)
        history["acc"].append(valid_acc)
        history["f1"].append(valid_f1)

        print(
            f"Epoch [{epoch + 1}/{epochs}] - step {step}, Train Loss: {average_train_loss}, "
            f"Val Loss: {valid_loss}, Val Acc:{valid_acc}, Val F1:{valid_f1}"
        )

        if round(best_acc, 5) < round(valid_acc, 5):
            best_acc = valid_acc
            best_acc_epoch = epoch
            if checkpoint_file is not None:
                model.save(
                    checkpoint_file,
                    train_dataloader.dataset.diacritics_vocab,
                    epoch,
                    valid_acc,
                    valid_f1,
                )
            nr_non_improving = 0
        else:
            nr_non_improving += 1
            print(f"Accuracy did not improve {best_acc} < {valid_acc}")

    if Path(checkpoint_file).exists():
        # Evaluate the existing model
        evaluate_step(1, 0, 0, 300_000 // valid_dataloader.batch_size)

    for epoch in range(epochs):
        model.train()

        step = 0
        running_loss = 0.0
        optimizer.zero_grad()

        epoch_predictions = []
        epoch_true_labels = []

        for (char_input, word_emb, sentence_emb), labels in train_dataloader:
            labels = labels.to(device)
            logits = model(
                char_input.to(device), word_emb.to(device), sentence_emb.to(device)
            )

            loss = loss_func(logits, labels)

            _, predicted = torch.max(logits, -1)
            epoch_predictions.extend(predicted.tolist())
            epoch_true_labels.extend([np.argmax(x) for x in labels.tolist()])

            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

            # update running values
            running_loss += loss.item()
            step += 1

            if step % 1000 == 0:
                epoch_accuracy = accuracy_score(epoch_true_labels, epoch_predictions)
                epoch_f1_metrics = f1_score(
                    np.array(epoch_true_labels).reshape(-1),
                    np.array(epoch_predictions).reshape(-1),
                    average="weighted",
                )
                average_train_loss = running_loss / step
                print(
                    f"Epoch {epoch+1} / Step: {step},
                    Loss: {loss.item()},
                    Avg Loss: {average_train_loss} ,
                    Train ACC: {epoch_accuracy},
                    Train F1: {epoch_f1_metrics}",
                )

            if step % 10000 == 0:
                evaluate_step(
                    step, epoch, running_loss, 300_000 // valid_dataloader.batch_size
                )
                model.train()
                if nr_non_improving >= patience:
                    print(f"Early stopping, patience ({patience}) ran out ")
                    break
        # print progress
        print(f"--------------------------- Epoch {epoch+1} -------------------------")
        evaluate_step(step, epoch, running_loss)

    if valid_dataloader is None and checkpoint_file is not None:
        model.save(checkpoint_file, train_dataloader.dataset.diacritics_vocab, epoch)

    print(f"Finished Training! Best Epoch = {best_acc_epoch}")


def evaluate(model, dataloader: DataLoader, loss_func, epoch=None, max_eval_steps=None):
    # print("***** Running prediction *****")
    model.eval()
    predict_out = []
    all_label_ids = []
    eval_loss = 0
    total = 0
    correct = 0
    steps = 0
    device = next(model.parameters()).device
    with torch.no_grad():
        for (char_input, word_emb, sentence_emb), labels in tqdm(
            dataloader, total=max_eval_steps
        ):
            labels = labels.to(device)
            logits = model(
                char_input.to(device), word_emb.to(device), sentence_emb.to(device)
            )

            loss = loss_func(logits, labels)

            eval_loss += loss.item()

            _, predicted = torch.max(logits, -1)

            predict_out.extend(predicted.tolist())
            all_label_ids.extend(labels.argmax(dim=1).tolist())
            eval_accuracy = predicted.eq(labels.argmax(dim=1)).sum().item()

            total += len(labels)
            correct += eval_accuracy
            steps += 1
            if max_eval_steps is not None and steps > max_eval_steps:
                break

        f1_metrics = f1_score(
            np.array(all_label_ids).reshape(-1),
            np.array(predict_out).reshape(-1),
            average="weighted",
        )
        report = classification_report(
            np.array(all_label_ids).reshape(-1),
            np.array(predict_out).reshape(-1),
            digits=4,
        )
        print("Evaluation Report")
        print(report)

    eval_acc = correct / total
    eval_loss = eval_loss / steps

    if epoch:
        print(
            f"Evaluation at Epoch: {epoch +1 }, Acc={eval_acc}, F1={f1_metrics}, Loss={eval_loss}"
        )
    else:
        print(f"Evaluation on Test, Acc={eval_acc}, F1={f1_metrics}, Loss={eval_loss}")

    return eval_loss, eval_acc, f1_metrics


def predict(model, dataloader: DataLoader):
    # print("***** Running prediction *****")
    model.eval()
    predict_out = []

    device = next(model.parameters()).device
    with torch.no_grad():
        for char_input, word_emb, sentence_emb in dataloader:
            logits = model(
                char_input.to(device), word_emb.to(device), sentence_emb.to(device)
            )

            predicted = F.softmax(logits, dim=1)

            predict_out.extend(predicted.tolist())

    return predict_out



from .diacritics_inference import restore_diacritics
from .diacritics_utils import (
    remove_diacritics,
    has_interesting_chars,
    correct_diacritics,
    DIACRITICS_CANDIDATES,
    MAP_DIACRITICS,
    LOG_NAME,
)



import re

LOG_NAME = "ro-diacritics"

MAP_DIACRITICS = {
    "ă": "a",
    "â": "a",
    "Â": "A",
    "Ă": "A",
    "ț": "t",
    "Ț": "T",
    "ș": "s",
    "Ș": "S",
    "î": "i",
    "Î": "I",
    "ş": "s",
    "Ş": "S",
    "ţ": "t",
    "Ţ": "T",
}
ALL_CANDIDATES = {
    "ă",
    "â",
    "a",
    "Ă",
    "Â",
    "A",
    "ț",
    "Ț",
    "ș",
    "Ș",
    "î",
    "Î",
    "i",
    "t",
    "s",
    "ş",
    "Ş",
    "ţ",
    "Ţ",
    "S",
    "T",
    "I",
}
DIACRITICS_CANDIDATES = {"a", "i", "s", "t"}
MAP_CORRECT_DIACRITICS = {
    "ş": "ș",
    "Ş": "Ș",
    "ţ": "ț",
    "Ţ": "Ț",
}
MAP_POSSIBLE_CHARS = {
    "a": ["ă", "â", "a"],
    "i": ["î", "i"],
    "s": ["ș", "s"],
    "t": ["ț", "t"],
}


def correct_diacritics(word):
    # use these three lines to do the replacement
    rep = dict((re.escape(k), v) for k, v in MAP_CORRECT_DIACRITICS.items())
    pattern = re.compile("|".join(rep.keys()))
    return pattern.sub(lambda m: rep[re.escape(m.group(0))], str(word))


def remove_diacritics(word):
    # use these three lines to do the replacement
    rep = dict((re.escape(k), v) for k, v in MAP_DIACRITICS.items())
    pattern = re.compile("|".join(rep.keys()))
    return pattern.sub(lambda m: rep[re.escape(m.group(0))], str(word))


def has_interesting_chars(word):
    # use these three lines to do the replacement
    pattern = re.compile("|".join(DIACRITICS_CANDIDATES))
    return pattern.search(str(word)) is not None



import zipfile
import logging
from pathlib import Path

import numpy as np
import torch
from torch.utils.data import DataLoader, IterableDataset
from collections import OrderedDict
from urllib.request import urlretrieve

from tqdm import tqdm

from .diacritics_dataset import DiacriticsDataset
from .diacritics_model import Diacritics
from .diacritcs_train import predict
from . import diacritics_dataset
from .diacritics_utils import LOG_NAME

logger = logging.getLogger(LOG_NAME)

_model: Diacritics = None
_dataset: DiacriticsDataset = None

CACHED_DIR = ".model"
MODEL_FILE = "diacritice.pt"
MODEL_URL = (
    "https://github.com/AndyTheFactory/andythefactory.github.io/raw/main/diacritice.zip"
)


def get_cached_model():
    """
    Loads the cached model. If not, it tries to download the model from github
    :return: local path to model
    """
    filename = Path(CACHED_DIR) / MODEL_FILE
    if filename.exists():
        return filename
    else:

        def reporthook(t):
            last_b = [0]

            def inner(b=1, bsize=1, tsize=None):
                if tsize is not None:
                    t.total = tsize
                t.update((b - last_b[0]) * bsize)
                last_b[0] = b

            return inner

        Path(CACHED_DIR).mkdir(exist_ok=True)
        dest_file = Path(CACHED_DIR) / Path(MODEL_URL).name
        with tqdm(unit="B", unit_scale=True, miniters=1, desc=str(filename)) as t:
            try:
                urlretrieve(MODEL_URL, dest_file, reporthook=reporthook(t))
            except KeyboardInterrupt as e:  # remove the partial zip file
                dest_file.unlink(missing_ok=True)
                raise e
        if dest_file.suffix == ".zip":
            with zipfile.ZipFile(dest_file, "r") as zf:
                zf.extractall(CACHED_DIR)
        return filename


def load_model(filename) -> (Diacritics, DiacriticsDataset):
    """
    Loads a trained :class:Diacritics model from cached file.
    Also, used hyperparams are loaded from the file
    :param filename: local path to stored model
    :return: loaded :class:Diacritics object and the used vocabulary (must be the same as in training)
    """
    import sys

    sys.modules["diacritics_dataset"] = diacritics_dataset

    checkpoint = torch.load(filename, map_location="cpu")
    params = checkpoint["hyperparams"]
    model = Diacritics(
        nr_classes=params["nr_classes"],
        word_embedding_size=params["word_embedding_size"],
        character_embedding_size=params["character_embedding_size"],
        char_vocabulary_size=params["char_vocabulary_size"],
        char_padding_index=params["char_padding_index"],
        character_window=params["character_window"],
        sentence_window=params["sentence_window"],
        characters_lstm_size=params["characters_lstm_size"],
        sentence_lstm_size=params["sentence_lstm_size"],
    )
    model.load_state_dict(checkpoint["model_state"])

    checkpoint["valid_f1"] = checkpoint["valid_f1"] if "valid_f1" in checkpoint else 0
    logger.info(
        f"Loaded checkpoint: Epoch: {checkpoint['epoch']}, valid_acc: {checkpoint['valid_acc']}, valid_f1: {checkpoint['valid_f1']}"
    )

    return model, checkpoint["vocabulary"]


def initmodel(filename=None):
    global _model, _dataset
    if filename is None:
        filename = get_cached_model()
    _model, vocab = load_model(filename)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if device == "cpu":
        logger.warning("GPU not available, using CPU")
    else:
        logger.info("GPU available")
    _model = _model.to(device)
    _dataset = DiacriticsDataset(
        "", _model.character_window, _model.sentence_window, diacritics_vocab=vocab
    )


def restore_diacritics(text, batch_size=128):
    """
    Transforms a Romanian text without diacritics (“ă”, “â”, “î”, “ș”, and “ț”) into a
    proper spelled text with romanian character set. The casing of the text is preserved.

    :param text: input text (without diacritics)
    :param batch_size: Adjust the batch size according to the available memory.
    :return: text with diacrtics replaced
    """
    if _model is None:
        initmodel()

    input_tensors, character_indices = _dataset.gen_batch(text, stride=10)

    class DS(IterableDataset):
        def __init__(self, input_tensors):
            self.input_tensors = input_tensors

        def __iter__(self):
            return iter(self.input_tensors)

    input_data = DataLoader(DS(input_tensors), batch_size=batch_size)

    predictions = predict(_model, input_data)

    prediction_tuples = sorted(zip(character_indices, predictions), key=lambda x: x[0])

    d = OrderedDict()

    for k, v in prediction_tuples:
        d.setdefault(k, []).append(v)

    text = list(text)
    for idx, p in d.items():
        p = np.argmax(np.average(p, axis=0))
        chr = DiacriticsDataset.get_char_from_label(text[idx].lower(), p)
        if text[idx].lower() == text[idx]:
            if chr == "â" and (
                idx == 0
                or idx == len(text) - 1
                or not text[idx - 1].isalpha()
                or not text[idx + 1].isalpha()
            ):
                #  â is not admitted at the beginning or end
                continue
            if chr == "î" and (
                idx > 0
                and idx < len(text) - 1
                and text[idx - 1].isalpha()
                and text[idx + 1].isalpha()
            ):
                # î is not admitted inside a word
                continue

            text[idx] = chr
        else:
            text[idx] = chr.upper()

    return "".join(text)



from setuptools import setup, find_packages

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setup(
    # $ pip install sampleproject
    name='ro-diacritics',  # Required

    version='0.9.4',  # Required

    description='Python API for Romanian diacritics restoration',  # Required

    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer="Andrei Paraschiv",
    maintainer_email="andrei@thephpfactory.com",
    author="Andrei Paraschiv",
    author_email="andrei@thephpfactory.com",

    url='https://github.com/AndyTheFactory/RO-Diacritics',  # Optional

    classifiers=[  # Optional
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',

        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Filters',
        'Topic :: Text Processing :: General',
        'Topic :: Text Processing :: Indexing',
        'Topic :: Text Processing :: Linguistic',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3'
    ],

    keywords='romanian diacritcs language restoration diacritice python',

    # package_dir={"": "."},
    packages=find_packages(),

    install_requires=['torch', 'torchtext', 'numpy', 'tqdm', 'nltk', 'scikit-learn',],

    zip_safe=False,
)



import pickle as pkl
from collections import Counter
from pathlib import Path

import numpy as np
import torch
from nltk.tokenize import TreebankWordTokenizer, PunktSentenceTokenizer
from nltk.tokenize import sent_tokenize, word_tokenize
from torch.utils.data import IterableDataset
from torchtext.vocab import FastText

from .diacritics_utils import (
    correct_diacritics,
    remove_diacritics,
    has_interesting_chars,
    DIACRITICS_CANDIDATES,
)


class DiacriticsVocab:
    def __init__(
        self,
        distinct_tokens: Counter,
        max_vocab,
        pad_token,
        unk_token,
        max_char_vocab,
        overflow_char,
    ):
        self.vocab = {
            "itos": {},
            "stoi": {},
            "vectors": torch.zeros(max_vocab + 2, 300, dtype=torch.float32),
        }
        self.pad_token = pad_token
        self.unk_token = unk_token
        self.max_char_vocab = max_char_vocab
        self.overflow_char = overflow_char

        embedding = FastText("ro")
        distinct_tokens = dict(distinct_tokens.most_common(max_vocab))

        self.vocab["itos"] = dict(enumerate(distinct_tokens, 1))
        self.vocab["itos"][self.pad_token] = "<pad>"
        self.vocab["itos"][self.unk_token] = "<unk>"

        self.vocab["stoi"] = {v: k for k, v in self.vocab["itos"].items()}

        fasttext_counts = Counter([remove_diacritics(word) for word in embedding.stoi])

        for (
            word,
            index,
        ) in embedding.stoi.items():  # Aggregate embeddings with diacritics
            word = remove_diacritics(word)
            if word in self.vocab["stoi"]:
                idx = self.vocab["stoi"][word]
                self.vocab["vectors"][idx] = (
                    self.vocab["vectors"][idx] + embedding.vectors[index]
                )

        for word, index in self.vocab["stoi"].items():
            if fasttext_counts[word] > 1:
                self.vocab["vectors"][index] = (
                    self.vocab["vectors"][index] / fasttext_counts[word]
                )

    def encode_char(self, c):
        return ord(c) if ord(c) <= self.max_char_vocab else self.overflow_char


class DiacriticsDataset(IterableDataset):
    def __init__(
        self,
        data,
        character_window,
        sentence_window,
        min_line_length=50,
        max_vocab=25000,
        max_char_vocab=770,
        overflow_char=255,
        diacritics_vocab: DiacriticsVocab = None,
    ):
        """

        :param data: Textfile, Pickle file or raw text
        :param character_window:
        :param sentence_window:
        :param min_line_length:
        :param max_vocab:
        :param max_char_vocab:
        :param overflow_char:
        :param diacritics_vocab:
        """
        self.character_window = character_window
        self.sentence_window = sentence_window
        self.min_line_length = min_line_length
        self.texts, distinct_tokens = self.load_texts(data, self.min_line_length)
        self.max_vocab = max_vocab
        self.pad_character = 0

        if diacritics_vocab is None:
            self.max_char_vocab = max_char_vocab
            self.overflow_char = overflow_char
            self.pad_token = 0
            self.unk_token = max_vocab + 1
        else:
            self.max_char_vocab = diacritics_vocab.max_char_vocab
            self.overflow_char = diacritics_vocab.overflow_char
            self.pad_token = diacritics_vocab.pad_token
            self.unk_token = diacritics_vocab.unk_token

        self.diacritics_vocab = (
            DiacriticsVocab(
                distinct_tokens,
                self.max_vocab,
                self.pad_token,
                self.unk_token,
                self.max_char_vocab,
                self.overflow_char,
            )
            if diacritics_vocab is None
            else diacritics_vocab
        )

    @property
    def vocab(self):
        return self.diacritics_vocab.vocab

    def encode_char(self, c):
        return self.diacritics_vocab.encode_char(c)

    def __iter__(self):
        return self.parse_text()

    @staticmethod
    def get_label(original_char):
        """
        :param original_char: lowercase diacritics char
        :return: 0 if no change (not turing into diacritics),
                 1 if changed to first candidate of diacritics
                 2 if changed to second candidate of diacritics (a has 2 candidates)
        """
        diacritic_to_label = {
            "ă": 1,
            "â": 2,
            "î": 1,
            "ș": 1,
            "ț": 1,
        }
        label = (
            diacritic_to_label[original_char]
            if original_char in diacritic_to_label
            else 0
        )
        label_tensor = torch.eye(3)

        return label_tensor[label]

    @staticmethod
    def get_char_from_label(original_char, label):
        if original_char not in DIACRITICS_CANDIDATES:
            return original_char

        if label == 1:
            return {
                "a": "ă",
                "i": "î",
                "s": "ș",
                "t": "ț",
            }[original_char]
        elif label == 2:
            return "â"
        else:
            return original_char

    def get_char_input(self, line, line_orig, token_idx):
        prefix_s = " ".join(line[:token_idx])
        suffix_s = " ".join(line[token_idx + 1 :])
        word = line[token_idx]
        word_orig = line_orig[token_idx].lower()
        encoded_list = []
        labels = []
        char_positions = []

        for ix, c in enumerate(word):
            if c in DIACRITICS_CANDIDATES:
                encode_text = (prefix_s + " " + word[:ix]).strip()
                encode_text = encode_text[-self.character_window :]
                l = (
                    [self.pad_character] * (self.character_window - len(encode_text))
                    + list(map(self.encode_char, encode_text))
                    + [self.encode_char(c)]
                )
                encode_text = (word[ix + 1 :] + " " + suffix_s).strip()
                encode_text = encode_text[: self.character_window]
                l = (
                    l
                    + list(map(self.encode_char, encode_text))
                    + [self.pad_character] * (self.character_window - len(encode_text))
                )
                encoded_list.append(l)
                labels.append(self.get_label(word_orig[ix]))
                char_positions.append(ix)
        return encoded_list, labels, char_positions

    def get_word_emb(self, word):
        if word in self.vocab["stoi"]:
            idx = self.vocab["stoi"][word]
            embed = self.vocab["vectors"][idx]
        else:
            embed = self.vocab["vectors"][self.unk_token]

        return embed

    def get_sentence_emb(self, sentence):
        encoded = list(map(self.get_word_emb, sentence))
        if len(encoded) < self.sentence_window:
            encoded = encoded + [self.get_word_emb("<pad>")] * (
                self.sentence_window - len(encoded)
            )
        return torch.stack(encoded)

    def parse_text(self):
        "Generates one sample of data"
        # Select sample
        for line_orig in self.texts:
            line = [
                remove_diacritics(x).lower() for x in line_orig
            ]  # embeddings based on words without diacritics
            line_indices = np.arange(len(line))
            if len(line_indices) > self.sentence_window:
                windows = np.lib.stride_tricks.sliding_window_view(
                    line_indices, self.sentence_window
                )
            else:
                windows = [line_indices]
            for window in windows:
                sentence_emb = self.get_sentence_emb(line[window[0] : window[-1] + 1])
                for token_idx in window:
                    if has_interesting_chars(line[token_idx]):
                        char_inputs, labels, _ = self.get_char_input(
                            line, line_orig, token_idx
                        )
                        word_emb = self.get_word_emb(line[token_idx])
                        for ix, char_input in enumerate(char_inputs):
                            yield (
                                torch.tensor(char_input),
                                word_emb,
                                sentence_emb,
                            ), labels[ix]

    def gen_batch(self, text, stride=1):
        """

        :param text: input text to be processed
        :param stride: generate sentence windows with this stride (you have to pool the results since you will
                        get more than 1 prediciton per character)
        :return:
        """
        text_plain = remove_diacritics(text).lower()
        lines = PunktSentenceTokenizer().span_tokenize(text)
        character_indices = []
        input_tensors = []

        for line_span in lines:
            line = text[line_span[0] : line_span[1]]
            words = list(TreebankWordTokenizer().span_tokenize(line))
            word_indices = np.arange(len(words))
            line_tokens = [line[x[0] : x[1]] for x in words]
            if len(words) > self.sentence_window:
                windows = np.lib.stride_tricks.sliding_window_view(
                    word_indices, self.sentence_window
                )[::stride, :]
                # add a reminder window (if it does not fit perfectly)
                if windows.max() < max(word_indices):
                    windows = np.vstack(
                        [windows, word_indices[-self.sentence_window :]]
                    )
            else:
                windows = [word_indices]

            for window in windows:
                sentence = [
                    line[x[0] : x[1]] for x in words[window[0] : window[-1] + 1]
                ]
                sentence_emb = self.get_sentence_emb(sentence)
                for token_idx in window:
                    word = line[words[token_idx][0] : words[token_idx][1]]
                    if has_interesting_chars(word):
                        char_inputs, _, char_positions = self.get_char_input(
                            line_tokens, line_tokens, token_idx
                        )
                        word_emb = self.get_word_emb(word)
                        for ix, char_input in enumerate(char_inputs):
                            input_tensors.append(
                                [torch.tensor(char_input), word_emb, sentence_emb]
                            )
                            character_indices.append(
                                line_span[0] + words[token_idx][0] + char_positions[ix]
                            )

        return input_tensors, character_indices

    @staticmethod
    def load_texts(data, min_line_length):
        if data and Path(data).exists():
            filename = data
            if Path(filename).suffix.lower() == ".pkl":  # loading cached pickle
                texts, distinct_no_diacritics = pkl.load(open(filename, "rb"))
            else:
                with open(filename, "r", encoding="utf-8") as f:
                    texts = [
                        correct_diacritics(line)
                        for line in f
                        if len(line) > min_line_length
                    ]

                texts = [
                    x
                    for line in texts
                    for x in sent_tokenize(line, "english")
                    if len(x) > min_line_length
                ]
                texts = [word_tokenize(line) for line in texts]

                distinct_no_diacritics = Counter(
                    [remove_diacritics(x) for line in texts for x in line]
                )
        else:
            texts = [
                x for x in sent_tokenize(data, "english") if len(x) > min_line_length
            ]
            texts = [word_tokenize(line) for line in texts]

            distinct_no_diacritics = Counter(
                [remove_diacritics(x) for line in texts for x in line]
            )

        return texts, distinct_no_diacritics
